/*
    Fichier smtp.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence.fr

    Traductions des droits d'utilisation dans les fichiers :
    licence.de , licence.en , licence.es , licence.it
    licence.nl , licence.pt , licence.eo , licence.eo-utf


    Bibliothque de fonctions permettant
    d'envoyer des mails au serveur smtp
*/

#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include "buflect.h"
#include "smtp.h"

//#define DEBUG

#define port_smtp     25  // port d'entre sortie pour les serveur smtp


/* variable globale au source
   (pour viter des tonnes de passages de paramtres) */

static int smtp = -1;     // descripteur pour dialoguer avec le serveur



/* connexion au serveur smtp */

int connect_smtp (char *serveur)
{
    struct hostent *HostEnt;      // description du host serveur
    struct sockaddr_in serv_addr; // addresse du serveur
    char bufw [80];               // buffer d'envoi d'une ligne ou requte smtp


    // rcupration adresse du serveur  partir de son nom
    HostEnt = gethostbyname (serveur);

    if (HostEnt == NULL)
    {
        // "Serveur smtp non trouv"
        affiche_err ("SERV_SMTP_ABSENT");
        return (0);
    }

    // initialisation structure serv_adr
    memset (&serv_addr, 0, sizeof (serv_addr));   // init serv_addr

    memcpy (&serv_addr.sin_addr, HostEnt->h_addr, HostEnt->h_length);

    serv_addr.sin_port = htons (port_smtp);  // host to network port
    serv_addr.sin_family = AF_INET;           // AF_*** : INET=internet

    // cration de la socket
    if ((smtp = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    {
        // "Problme cration socket client smtp"
        affiche_err ("CREAT_SOCKET_SMTP");
        return (0);
    }

    // requte de connexion
    if (connect (smtp, (struct sockaddr*) &serv_addr, sizeof (serv_addr)) < 0)
    {
        // "Problme demande de connexion"
        // "peut tre due  une connexion Internet par proxy"
        affiche_err ("PB_ACCES_SERV-1");
        affiche_err ("PB_ACCES_SERV-2");
        return (0);
    }
    else
    {
        // attente rponse serveur smtp
        lire_smtp (smtp);

        // gnrer le message  helo nom_de_calculateur
        strcpy (bufw, "HELO ");
        gethostname (bufw + 5, sizeof (bufw) - 5);

        // identification du calculateur connect
        envoie_smtp (bufw);
        lire_smtp ();
        return (1);
    }
}



/* Ecriture de donnes
   envoie la chaine passe en paramtre suivie de \r\n
*/

void envoie_smtp (char *buffer)
{
#ifdef DEBUG
    putchar ('>');
    puts (buffer);
    fflush (stdout);
#endif

    write (smtp, buffer, strlen (buffer));
    write (smtp, "\r\n", 2);
}



/* Lecture d'une ligne de donnes
   la lecture s'arrte sur un caractre de passage  la ligne
*/

void lire_smtp ()
{
    int posbuf;


    // initialisation
    posbuf = 0;

    // lecture
    do
        read (smtp, buf_lect + posbuf, 1);
    while (buf_lect [posbuf++] != '\n' && posbuf < sz_buflect);

#ifdef DEBUG
    buf_lect [posbuf - 1] = '\0';
    putchar ('<');
    puts (buf_lect);
    fflush (stdout);
#endif
}



/* dconnexion propre du serveur smtp */

void deconnect_smtp ()
{
    // envoi du message de deconnexion
    envoie_smtp ("QUIT");

    // lecture de l'acquittement
    lire_smtp ();

    // fermeture de la connexion
    shutdown (smtp, 2);
    close (smtp);
}
